
import torch
from torch import nn
from torch.utils.data import DataLoader
from torchvision import datasets
from torchvision.transforms import ToTensor
from mpl_toolkits.axes_grid1 import ImageGrid
from torchvision.utils import save_image, make_grid
import matplotlib.pyplot as plt

# Download training data from open datasets.
training_data = datasets.MNIST(
	root="data",
	train=True,
	download=False,
	transform=ToTensor(),
)

# Download test data from open datasets.
test_data = datasets.MNIST(
	root="data",
	train=False,
	download=False,
	transform=ToTensor(),
)


batch_size = 100

# Create data loaders.
# train_dataloader = DataLoader(training_data, batch_size=batch_size, shuffle=True )
train_dataloader = DataLoader(training_data, batch_size=batch_size )
test_dataloader = DataLoader(test_data, batch_size=batch_size, shuffle=False )

# device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
device = torch.device("mps" if torch.cuda.is_available() else "cpu")


class VAE(nn.Module):
	def __init__(self, input_dim=784, hidden_dim=400, latent_dim=200, device=device):
		super(VAE, self).__init__()

		# encoder
		self.encoder = nn.Sequential(
			nn.Linear(input_dim, hidden_dim),
			nn.LeakyReLU(0.2),
			nn.Linear(hidden_dim, latent_dim),
			nn.LeakyReLU(0.2)
			)
		
		# latent mean and variance 
		self.mean_layer = nn.Linear(latent_dim, 2)
		self.logvar_layer = nn.Linear(latent_dim, 2)
		
		# decoder
		self.decoder = nn.Sequential(
			nn.Linear(2, latent_dim),
			nn.LeakyReLU(0.2),
			nn.Linear(latent_dim, hidden_dim),
			nn.LeakyReLU(0.2),
			nn.Linear(hidden_dim, input_dim),
			nn.Sigmoid()
			)
	 
	def encode(self, x):
		x = self.encoder(x)
		mean, logvar = self.mean_layer(x), self.logvar_layer(x)
		return mean, logvar

	def reparameterization(self, mean, var):
		epsilon = torch.randn_like(var).to(device)	  
		z = mean + var*epsilon
		return z

	def decode(self, x):
		return self.decoder(x)

	def forward(self, x):
		mean, logvar = self.encode(x)
		z = self.reparameterization(mean, logvar)
		x_hat = self.decode(z)
		return x_hat, mean, logvar


model = VAE().to(device)
optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)

def loss_function(x, x_hat, mean, log_var):
	reproduction_loss = nn.functional.binary_cross_entropy(x_hat, x, reduction='sum')
	KLD = - 0.5 * torch.sum(1+ log_var - mean.pow(2) - log_var.exp())

	return reproduction_loss + KLD


def train(model, optimizer, epochs, device, x_dim=784):
	model.train()
	for epoch in range(epochs):
		overall_loss = 0
		for batch_idx, (x, _) in enumerate(train_dataloader):
			x = x.view(batch_size, x_dim).to(device)

			optimizer.zero_grad()

			x_hat, mean, log_var = model(x)
			loss = loss_function(x, x_hat, mean, log_var)
			
			overall_loss += loss.item()
			
			loss.backward()
			optimizer.step()

		print("\tEpoch", epoch + 1, "\tAverage Loss: ", overall_loss/(batch_idx*batch_size))
	return overall_loss

train(model, optimizer, epochs=50, device=device)

torch.save( model, "vae1.pth" )

