import torch
from torch import nn
from torch.utils.data import DataLoader
from torchvision import datasets
from torchvision.transforms import ToTensor
from mpl_toolkits.axes_grid1 import ImageGrid
from torchvision.utils import save_image, make_grid
import matplotlib.pyplot as plt

# Download training data from open datasets.
training_data = datasets.MNIST(
	root="data",
	train=True,
	download=False,
	transform=ToTensor(),
)

# Download test data from open datasets.
test_data = datasets.MNIST(
	root="data",
	train=False,
	download=False,
	transform=ToTensor(),
)


batch_size = 100

# Create data loaders.
# train_dataloader = DataLoader(training_data, batch_size=batch_size, shuffle=True )
train_dataloader = DataLoader(training_data, batch_size=batch_size )
test_dataloader = DataLoader(test_data, batch_size=batch_size, shuffle=False )

# device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
device = torch.device("mps" if torch.cuda.is_available() else "cpu")


class VAE(nn.Module):
	def __init__(self, input_dim=784, hidden_dim=400, latent_dim=200, device=device):
		super(VAE, self).__init__()

		# encoder
		self.encoder = nn.Sequential(
			nn.Linear(input_dim, hidden_dim),
			nn.LeakyReLU(0.2),
			nn.Linear(hidden_dim, latent_dim),
			nn.LeakyReLU(0.2)
			)
		
		# latent mean and variance 
		self.mean_layer = nn.Linear(latent_dim, 2)
		self.logvar_layer = nn.Linear(latent_dim, 2)
		
		# decoder
		self.decoder = nn.Sequential(
			nn.Linear(2, latent_dim),
			nn.LeakyReLU(0.2),
			nn.Linear(latent_dim, hidden_dim),
			nn.LeakyReLU(0.2),
			nn.Linear(hidden_dim, input_dim),
			nn.Sigmoid()
			)
	 
	def encode(self, x):
		x = self.encoder(x)
		mean, logvar = self.mean_layer(x), self.logvar_layer(x)
		return mean, logvar

	def reparameterization(self, mean, var):
		epsilon = torch.randn_like(var).to(device)	  
		z = mean + var*epsilon
		return z

	def decode(self, x):
		return self.decoder(x)

	def forward(self, x):
		mean, logvar = self.encode(x)
		z = self.reparameterization(mean, logvar)
		x_hat = self.decode(z)
		return x_hat, mean, logvar

model = torch.load('vae1.pth', weights_only=False)
model.eval( )


def generate_digit(mean, var):
	z_sample = torch.tensor([[mean, var]], dtype=torch.float).to(device)
	x_decoded = model.decode(z_sample)
	digit = x_decoded.detach().cpu().reshape(28, 28) # reshape vector to 2d array
	plt.title(f'[{mean},{var}]')
	plt.imshow(digit, cmap='gray')
	plt.axis('off')
	plt.show()
	 

#img1: mean0, var1 / img2: mean1, var0
generate_digit(0.0, 1.0), generate_digit(1.0, 0.0)
	 
